/* works2txt.c */

/* Convert comma separated files from Microsoft Works to a form that can
	be read by txt2siag. Problem: MS Works uses commas to separate
	the fields *and* to separate the thousands from the millions
	inside numbers. To distinguish between the cases, Works puts
	"" around the numbers. Another problem is that Works puts ""
	around any label. To solve these problems, every field must
	be analyzed and "" removed. Furthermore, remove ',' that appear
	inside a field. This becomes a problem when a ',' appears
	inside a label, but we'll just have to live with that for now.

	To accomplish this, use a state machine with these states:
	START: if nextc is " delete and go to INSIDE, 
		if nextc is , or \n goto START
		else goto OUTSIDE
	INSIDE: if nextc is , delete and go to INSIDE
		if nextc is " delete and go to OUTSIDE
		if nextc is \n go to START
		else goto INSIDE
	OUTSIDE: if nextc is , or \n goto START
		else go to OUTSIDE
*/

#include <stdio.h>

#define START 0
#define INSIDE 1
#define OUTSIDE 2


int
main(int argc, char **argv)
{
	int state = START;
	int nextc;

	while ((nextc = getchar()) != EOF) {
		if (nextc == '\n') {
			putchar(nextc);
			state = START;
		} else if (isspace(nextc)) {
			putchar(nextc);
		} else {
			switch (state) {
			case START:
				if (nextc == '\"') state = INSIDE;
				else if (nextc == ',')
					putchar(nextc);
				else {
					putchar(nextc);
					state = OUTSIDE;
				}
				break;
			case INSIDE:
				if (nextc == ',');
				else if (nextc == '\"') state = OUTSIDE;
				else putchar(nextc);
				break;
			case OUTSIDE:
				if (nextc == ',') state = START;
				putchar(nextc);
				break;
			}
		}
	}
}

