/*
   Siag, Scheme In A Grid
   Copyright (C) 1996  Ulric Eriksson <ulric@asogy.stockholm.se>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 1, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
   Module name:    window.c

   This module handles windows: creating, destroying, printing on windows.
*/

#include <stdio.h>
#include <string.h>
#include <malloc.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/cursorfont.h>

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Scrollbar.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Command.h>

#include "types.h"
#include "calc.h"

#include "../xstuff/fonts.h"

window *w_list;

int pr_scr_flag;

Display *display;

Window root/*, cell_win*/;
static GC cursor_gc, colnum_gc, rownum_gc, cell_gc;

void draw_input(Display * display, char *text)
{
	char b[256];
	sprintf(b, "[%d,%d] %s",
		w_list->point_pos.row, w_list->point_pos.col, text);
	XtVaSetValues(label1, XtNlabel, b, NULL);
}

void draw_status(Display * display, char *text)
{
	XtVaSetValues(label2, XtNlabel, text, NULL);
	XFlush(display);
}

/*
   void llpr(char *p)
   Prints the string p on the bottom line of the screen.  If p is empty and
   the last string printed was also empty, the string isn't printed.
*/
void llpr(char *p)
{
	static int isclear = FALSE;

	if (isclear && p[0] == '\0')
		return;
	isclear = (p[0] == '\0');

	draw_status(display, p);
}

void draw_vbar(Display * display, window *w)
{
	Arg arg;
	union {
		int int_value;
		float float_value;
	} value;

	/* this is voodoo */
	value.float_value = w->top.row;
	value.float_value /= BUFFER_ROWS;
	XtSetArg(arg, XtNtopOfThumb, value.int_value);
	XtSetValues(w->vscroll, &arg, 1);
}

void draw_hbar(Display * display, window *w)
{
	Arg arg;
	union {
		int int_value;
		float float_value;
	} value;

	/* this is voodoo */
	value.float_value = w->top.col;
	value.float_value /= BUFFER_COLS;
	XtSetArg(arg, XtNtopOfThumb, value.int_value);
	XtSetValues(w->hscroll, &arg, 1);
}

void draw_scrollbars(Display * display, window *w)
{
	draw_vbar(display, w);
	draw_hbar(display, w);
}


void draw_colnums(Display * display, window *w)
{
	int x, y, i, x_pos, y_pos, x_base;
	char buf[80];
	unsigned int width, height, border, depth, text_height;

	XGetGeometry(display, XtWindow(w->colnum), &root,
		     &x, &y, &width, &height, &border, &depth);
	text_height = font_height(display, COLNUM_FONT);
	y_pos = (height - text_height) / 2 + font_ascent(display, COLNUM_FONT);
	XClearWindow(display, XtWindow(w->colnum));
	x_base = 0;
	x_pos = 5;
	for (i = w->top.col; x_base + x_pos < width; i++) {
		if (i > max_columns) break;
		sprintf(buf, "%d", i);
		XDrawString(display, XtWindow(w->colnum), colnum_gc,
			x_base + x_pos, y_pos, buf, strlen(buf));
		XDrawLine(display, XtWindow(w->colnum), colnum_gc,
			  x_base + cell_width(w->buf, i), 20,
			  x_base + cell_width(w->buf, i), 0);
		x_base += cell_width(w->buf, i);
	}
}

void draw_rownums(Display * display, window *w)
{
	int x, y, i, x_pos, y_pos, y_base;
	char buf[80];
	unsigned int width, height, border, depth, text_height;

	XGetGeometry(display, XtWindow(w->rownum), &root,
		     &x, &y, &width, &height, &border, &depth);
	text_height = font_height(display, ROWNUM_FONT);
	XClearWindow(display, XtWindow(w->rownum));
	x_pos = 5;
	y_pos = (cell_height(w->buf, w->top.row) - text_height) / 2
		+ font_ascent(display, ROWNUM_FONT);
	y_base = 0;
	for (i = w->top.row; y_base + y_pos < height; i++) {
		if (i > max_lines) break;

		sprintf(buf, "%d", i);
		y_pos = (cell_height(w->buf, i) - text_height) / 2
			+ font_ascent(display, ROWNUM_FONT);
		XDrawString(display, XtWindow(w->rownum), rownum_gc,
			x_pos, y_base + y_pos, buf, strlen(buf));
		XDrawLine(display, XtWindow(w->rownum), rownum_gc, 0,
			  y_base + cell_height(w->buf, i),
			  80, y_base + cell_height(w->buf, i));
		y_base += cell_height(w->buf, i);
	}
}

static int inblock(window *w, int r, int c)
{
	return ((r >= w->blku.row) && (r <= w->blkl.row) &&
		(c >= w->blku.col) && (c <= w->blkl.col));
}				/* inblock */

void draw_cell(Display *display, Window cell_win, buffer *buf,
	       int row, int col, int x_base, int y_base, window *w)
{
	int font_index;
	Font font;
	char b[80];
	int x_pos, y_pos, f;
	unsigned int text_height;

	if (row > max_lines || col > max_columns) return;

	font_index = ret_font(buf->matrix, row, col);
	font = get_font(display, font_index);
	XSetFont(display, cell_gc, font);
	text_height = font_height(display, font_index);

	x_pos = 5;
	y_pos = (cell_height(buf, row) + text_height) / 2
		- font_descent(display, font_index);
	b[0] = '\0';
	switch (ret_type(buf->matrix, row, col)) {
	case LABEL:
		sprintf(b, "%s", ret_taddr(buf->matrix, row, col));
		break;
	case CONSTANT:
	case EXPRESSION:
		sprintf(b, "%.2f", ret_val(buf->matrix, row, col));
		break;
	default:
		break;
	}
	XDrawString(display, cell_win, cell_gc, x_base + x_pos, y_base + y_pos,
		    b, strlen(b));
	/* These lines mark the cell borders */
	XDrawLine(display, cell_win, cell_gc,
		  x_base, y_base + cell_height(buf, row),
	  x_base + cell_width(buf, col), y_base + cell_height(buf, row));
	XDrawLine(display, cell_win, cell_gc,
	   x_base + cell_width(buf, col), y_base + cell_height(buf, row),
		  x_base + cell_width(buf, col), y_base);
	/* These lines are from the format */
	f = ret_format(buf->matrix, row, col);
	if (f & BORDER_LEFT)
		XDrawLine(display, cell_win, cell_gc,
			  x_base + 1, y_base + 1,
			  x_base + 1, y_base + cell_height(buf, row) - 1);
	if (f & BORDER_RIGHT)
		XDrawLine(display, cell_win, cell_gc,
			  x_base + cell_width(buf, col) - 1, y_base + 1,
			  x_base + cell_width(buf, col) - 1, y_base + cell_height(buf, row) - 1);
	if (f & BORDER_TOP)
		XDrawLine(display, cell_win, cell_gc,
			  x_base + 1, y_base + 1,
			  x_base + cell_width(buf, col) - 1, y_base + 1);
	if (f & BORDER_BOTTOM)
		XDrawLine(display, cell_win, cell_gc,
			  x_base + 1, y_base + cell_height(buf, row) - 1,
			  x_base + cell_width(buf, col) - 1, y_base + cell_height(buf, row) - 1);
	if (inblock(w, row, col)) {
		XFillRectangle(display, cell_win, cursor_gc, x_base, y_base,
			    cell_width(buf, col), cell_height(buf, row));
	}
}

void draw_cells(Display * display, window *w)
{
	int x, y, i, j, x_base, y_base;
	unsigned int width, height, border, depth;
	Window cell_win = XtWindow(w->grid);

	XGetGeometry(display, cell_win, &root,
		     &x, &y, &width, &height, &border, &depth);

	y_base = 0;
	XClearWindow(display, cell_win);
	for (i = w->top.row; y_base < height; i++) {
		x_base = 0;
		for (j = w->top.col; x_base < width; j++) {
			draw_cell(display, cell_win, w->buf,
				  i, j, x_base, y_base, w);
			x_base += cell_width(w->buf, j);
		}
		y_base += cell_height(w->buf, i);
	}
}

GC
get_gc(Display * display, Window wind,
       unsigned long fg, unsigned long bg, Font font)
{
	unsigned long valuemask = 0/*GCFunction | GCForeground | GCBackground |
		GCLineWidth | GCFont*/;
	XGCValues values;
	GC gc = XCreateGC(display, wind, valuemask, &values);

	XSetForeground(display, gc, fg);
	XSetBackground(display, gc, bg);
	if (font != -1)
		XSetFont(display, gc, font);
	return gc;
}

window *
find_window_by_widget(Widget wdg)
{
	window *w = w_list;
	do {
		if (w->viewport == wdg || w->selectall == wdg ||
			w->colnum == wdg || w->rownum == wdg ||
			w->grid == wdg || w->vscroll == wdg ||
			w->hscroll == wdg)
			return w;
		w = w->next;
	} while (w != w_list);
	return NULL;
}

void
free_window(window *w)
{
	window *pw;

	for (pw = w_list; pw->next != w && pw->next != pw; pw = pw->next);
	pw->next = w->next;

	if (w_list == w) w_list = w_list->next;
	if (w_list == w) w_list = NULL;
	XtDestroyWidget(w->viewport);
	free(w);
}

static int width(Widget w)
{
	Dimension x;

	XtVaGetValues(w, XtNwidth, &x, NULL);
	return x;
}

static int height(Widget w)
{
	Dimension x;

	XtVaGetValues(w, XtNheight, &x, NULL);
	return x;
}

window *
new_window(buffer *b, window *prev)
{
	Dimension totalwidth, formheight, w1, h1;
	int d;	/* distance between widgets in form */
	window *w;

	w = (window *)malloc(sizeof(window));

	if (w == NULL) return NULL;

	w->buf = b;
	w->point_pos.row = 1;
	w->point_pos.col = 1;
	w->top = w->point_pos;

	if (prev == NULL) prev = w;
	else w->next = prev->next;
	prev->next = w;

	/* Figure out how big the new form should be.			*/
	/* The total width must be the width of the gridpane.		*/
	XtVaGetValues(gridpane,
		XtNwidth, &totalwidth, NULL);

#ifdef DEBUG
	printf("Total width: %d\n", totalwidth);
#endif

	/* The form height is whatever we get, but if it is too small
	we cannot create the new window.				*/
	formheight = 100;
	w->viewport = XtVaCreateManagedWidget("viewport",
		formWidgetClass, gridpane,
		XtNwidth, totalwidth,
		/*XtNheight, formheight, */NULL);
	XtVaGetValues(w->viewport,
		XtNheight, &formheight, XtNdefaultDistance, &d, NULL);

#ifdef DEBUG
	printf("Form height: %d\n", formheight);
#endif

	/* The selectall button should get its size from the ad */
	w->selectall = XtVaCreateManagedWidget("selectall",
		commandWidgetClass, w->viewport, NULL);

	/* The colnum should take up all the space between the button
	and vscroll							*/
	w1 = totalwidth-width(w->selectall)-20-4*d;	/* just guessing */
	h1 = height(w->selectall);

#ifdef DEBUG
	printf("colnum size: w=%d h=%d\n", w1, h1);
#endif
	w->colnum = XtVaCreateManagedWidget("colnum",
		coreWidgetClass, w->viewport,
		XtNwidth, w1, XtNheight, h1,
		XtNfromHoriz, w->selectall, NULL);
	w1 = width(w->selectall);
	h1 = formheight-height(w->selectall)-20-4*d;

#ifdef DEBUG
	printf("rownum size: w=%d h=%d\n", w1, h1);
#endif

	w->rownum = XtVaCreateManagedWidget("rownum",
		coreWidgetClass, w->viewport,
		XtNwidth, w1, XtNheight, h1,
		XtNfromVert, w->selectall, NULL);
	w1 = width(w->colnum);
	h1 = height(w->rownum);

#ifdef DEBUG
	printf("grid size: w=%d h=%d\n", w1, h1);
#endif

	w->grid = XtVaCreateManagedWidget("grid",
		coreWidgetClass, w->viewport,
		XtNwidth, w1, XtNheight, h1,
		XtNfromHoriz, w->selectall,
		XtNfromVert, w->selectall, NULL);
	h1 = height(w->colnum)+height(w->grid)+d;
	w->vscroll = XtVaCreateManagedWidget("vscroll",
		scrollbarWidgetClass, w->viewport,
		XtNheight, h1,
		XtNfromHoriz, w->colnum, NULL);
	w1 = width(w->rownum)+width(w->grid)+d;
	w->hscroll = XtVaCreateManagedWidget("hscroll",
		scrollbarWidgetClass, w->viewport,
		XtNwidth, w1,
		XtNfromVert, w->rownum, NULL);
	XtAddCallback(w->vscroll, XtNjumpProc, vscroll_jump, NULL);
	XtAddCallback(w->vscroll, XtNscrollProc, vscroll_scroll, NULL);

	XtAddCallback(w->hscroll, XtNjumpProc, hscroll_jump, NULL);
	XtAddCallback(w->hscroll, XtNscrollProc, hscroll_scroll, NULL);

	XtVaGetValues(w->viewport,
		XtNheight, &formheight, NULL);

#ifdef DEBUG
	printf("Form height: %d\n", formheight);
#endif

	return w;
}

int
remove_window(window *w)
{
	
	if (w == w->next) return FALSE;
	free_window(w);
	return TRUE;
}

int
split_window(window *w)
{
	window *w2 = new_window(w->buf, w);

	if (w2 == NULL) return FALSE;
	w2->point_pos = w->point_pos;
	w2->top = w->top;
	return TRUE;
}

/*
   void init_windows(buffer *b)
   Sets up the whole initial window structure and initializes scrupd.
   The window list w_list is set to a list with a single window with the
   buffer b.
*/
void init_windows(buffer * b, int argc, char **argv)
{
	char *p;

	unsigned long foreground;	/* pixel value */
	unsigned long background;
	Window cell_win;

	w_list = new_window(b, NULL);

	XtRealizeWidget(topLevel);
#if 0
	w_list = malloc(sizeof(window));
	w_list->buf = b;
	w_list->point_pos.row = 1;
	w_list->point_pos.col = 1;
	w_list->top = w_list->point_pos;
#endif
	display = XtDisplay(topLevel);

	root = DefaultRootWindow(display);

	foreground = BlackPixelOfScreen(XtScreen(topLevel));
	cell_win = XtWindow(w_list->grid);

	XtVaGetValues(w_list->colnum,
		XtNbackground, &background, NULL);
	colnum_gc = get_gc(display, XtWindow(w_list->colnum), foreground, background,
			   get_font(display, COLNUM_FONT));
	XtVaGetValues(w_list->rownum,
		XtNbackground, &background, NULL);
	rownum_gc = get_gc(display, XtWindow(w_list->rownum), foreground, background,
			   get_font(display, ROWNUM_FONT));
	XtVaGetValues(w_list->grid,
		XtNbackground, &background, NULL);
	cell_gc = get_gc(display, cell_win, foreground, background,
			 get_font(display, CELL_FONT));
	cursor_gc = get_gc(display, cell_win, foreground ^ background, 0, -1);
	XSetFunction(display, cursor_gc, GXxor);
	XSetLineAttributes(display, cursor_gc, 2, LineSolid, CapButt, JoinMiter);

	draw_scrollbars(display, w_list);
	p = ret_taddr(w_list->buf->matrix,
		      w_list->point_pos.row, w_list->point_pos.col);
	if (p == NULL)
		p = "";
	draw_input(display, p);
	draw_status(display, "");
	draw_colnums(display, w_list);
	draw_rownums(display, w_list);
	draw_cells(display, w_list);
}

/*
   void exit_windows()
   Cleans up after Calc before exit.  All buffers and windows are freed.
*/
void exit_windows()
{
	/* free all buffers */
	while (b_list != NULL)
		free_buffer(b_list);
	while (w_list != NULL)
		free_window(w_list);
}

/*
   static void pr_scr()
   Prints and refreshes all the windows.
   Sets pr_scr_flag to FALSE.
*/
static void pr_scr()
{
	window *w;

	draw_status(display, "");
	w = w_list;
	do {
		draw_colnums(display, w);
		draw_rownums(display, w);
		draw_cells(display, w);
		draw_scrollbars(display, w);
		w = w->next;
	} while (w != w_list);
	pr_scr_flag = FALSE;
}	/* pr_scr */

void get_cell_coords(window *w, int top_row, int top_col,
		     int cell_row, int cell_col,
		     int *cell_x, int *cell_y)
{
	*cell_y = 0;
	while (cell_row < top_row) {
		cell_row++;
		*cell_y -= cell_height(w->buf, cell_row);
	}
	while (cell_row > top_row) {
		cell_row--;
		*cell_y += cell_height(w->buf, cell_row);
	}
	*cell_x = 0;
	while (cell_col < top_col) {
		cell_col++;
		*cell_x -= cell_width(w->buf, cell_col);
	}
	while (cell_col > top_col) {
		cell_col--;
		*cell_x += cell_width(w->buf, cell_col);
	}
}

void get_coords_cell(window *w, int top_row, int top_col,
		     int *cur_row, int *cur_col,
		     int cur_x, int cur_y)
{
	*cur_row = top_row;
	*cur_col = top_col;
	while (cur_y < 0) {
		cur_y += cell_height(w->buf, *cur_row);
		(*cur_row)--;
	}
	while (cur_y > cell_height(w->buf, *cur_row)) {
		cur_y -= cell_height(w->buf, *cur_row);
		(*cur_row)++;
	}
	while (cur_x < 0) {
		cur_x += cell_width(w->buf, *cur_col);
		(*cur_col)--;
	}
	while (cur_x > cell_width(w->buf, *cur_col)) {
		cur_x -= cell_width(w->buf, *cur_col);
		(*cur_col)++;
	}
}

/*
   void show_cur(window *w)
   Moves the cursor to reflect the position of point in w.
   If point is not visible, the window is moved so that point is in
   the middle of the screen.
*/
void show_cur(window *w)
{
	Window root;
	int x, y, cur_x, cur_y;
	unsigned int width, height, border_width, depth;
	char *p;
	Window cell_win = XtWindow(w->grid);

	/* Figure out how big the window is */
	XGetGeometry(display, cell_win, &root, &x, &y, &width, &height,
		     &border_width, &depth);
	get_cell_coords(w, w->top.row, w->top.col,
			w->point_pos.row, w->point_pos.col,
			&cur_x, &cur_y);
	/* this isn't efficient, but it will work */
	while (cur_y < 0) {
		cur_y += cell_height(w->buf, w->top.row);
		w->top.row--;
		pr_scr_flag = TRUE;
	}
	while (cur_y > height - cell_height(w->buf, w->top.row)) {
		cur_y -= cell_height(w->buf, w->top.row);
		w->top.row++;
		pr_scr_flag = TRUE;
	}
	while (cur_x < 0) {
		cur_x += cell_width(w->buf, w->top.col);
		w->top.col--;
		pr_scr_flag = TRUE;
	}
	while (cur_x > width - cell_width(w->buf, w->top.col)) {
		cur_x -= cell_width(w->buf, w->top.col);
		w->top.col++;
		pr_scr_flag = TRUE;
	}
	if (pr_scr_flag) {
		pr_scr();
	}
	p = ret_taddr(w->buf->matrix, w->point_pos.row,
		      w->point_pos.col);
	if (p == NULL)
		p = "";
	draw_input(display, p);
	XDrawRectangle(display, cell_win, cursor_gc, cur_x + 3, cur_y + 3,
		       cell_width(w->buf, w->point_pos.col) - 5,
		    cell_height(w->buf, w->point_pos.row) - 5);
}	/* show_cur */

void hide_cur(window *w)
{
	int cur_x, cur_y;
	Window cell_win = XtWindow(w->grid);

	/* here we don't really care if the cursor is in the window or not */
	get_cell_coords(w, w->top.row, w->top.col,
			w->point_pos.row, w->point_pos.col,
			&cur_x, &cur_y);
	XDrawRectangle(display, cell_win, cursor_gc, cur_x + 3, cur_y + 3,
		       cell_width(w->buf, w->point_pos.col) - 5,
		    cell_height(w->buf, w->point_pos.row) - 5);
}

